;;; Commentary:

;;; This is the famous Hello-World-program, written for Guile.  
;;;
;;; For an advanced version, see the script `hello' in the same
;;; directory.

;;; Author: Martin Grabmueller
;;; Date: 2001-05-29

;;; Code:

(display "Hello, World!")
(newline)

;;; End of file.
